package org.jboss.cache.cluster;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.RPCManager;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.factories.annotations.Inject;
import org.jboss.cache.factories.annotations.Start;
import org.jboss.cache.factories.annotations.Stop;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodCallFactory;
import org.jboss.cache.marshall.MethodDeclarations;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

/**
 * Periodically (or when certain size is exceeded) takes elements and replicates them.
 *
 * @author <a href="mailto:bela@jboss.org">Bela Ban</a> May 24, 2003
 * @version $Revision: 5197 $
 */
public class ReplicationQueue
{

   private static Log log = LogFactory.getLog(ReplicationQueue.class);

   /**
    * We flush every 5 seconds. Inactive if -1 or 0
    */
   private long interval = 5000;

   /**
    * Max elements before we flush
    */
   private long max_elements = 500;

   /**
    * Holds the replication jobs: LinkedList<MethodCall>
    */
   final List<MethodCall> elements = new LinkedList<MethodCall>();

   /**
    * For periodical replication
    */
   private Timer timer = null;

   /**
    * The timer task, only calls flush() when executed by Timer
    */
   private ReplicationQueue.MyTask task = null;
   private RPCManager rpcManager;
   private Configuration configuration;
   private boolean enabled;


   public boolean isEnabled()
   {
      return enabled;
   }

   @Inject
   private void injectDependencies(RPCManager rpcManager, Configuration configuration)
   {
      this.rpcManager = rpcManager;
      this.configuration = configuration;
      enabled = configuration.isUseReplQueue() && (configuration.getBuddyReplicationConfig() == null || !configuration.getBuddyReplicationConfig().isEnabled());
   }

   /**
    * Starts the asynchronous flush queue.
    */
   @Start
   public synchronized void start()
   {
      this.interval = configuration.getReplQueueInterval();
      this.max_elements = configuration.getReplQueueMaxElements();
      // check again
      enabled = configuration.isUseReplQueue() && (configuration.getBuddyReplicationConfig() == null || !configuration.getBuddyReplicationConfig().isEnabled());
      if (enabled)
      {
         if (interval > 0)
         {
            if (task == null)
               task = new ReplicationQueue.MyTask();
            if (timer == null)
            {
               timer = new Timer(true);
               timer.schedule(task,
                     500, // delay before initial flush
                     interval); // interval between flushes
            }
         }
      }
   }

   /**
    * Stops the asynchronous flush queue.
    */
   @Stop
   public synchronized void stop()
   {
      if (task != null)
      {
         task.cancel();
         task = null;
      }
      if (timer != null)
      {
         timer.cancel();
         timer = null;
      }
   }


   /**
    * Adds a new method call.
    */
   public void add(MethodCall job)
   {
      if (job == null)
         throw new NullPointerException("job is null");
      synchronized (elements)
      {
         elements.add(job);
         if (elements.size() >= max_elements)
            flush();
      }
   }

   /**
    * Flushes existing method calls.
    */
   public void flush()
   {
      List<MethodCall> l;
      synchronized (elements)
      {
         if (log.isTraceEnabled())
            log.trace("flush(): flushing repl queue (num elements=" + elements.size() + ")");
         l = new ArrayList<MethodCall>(elements);
         elements.clear();
      }

      if (l.size() > 0)
      {
         try
         {
            // send to all live nodes in the cluster
            rpcManager.callRemoteMethods(null, MethodCallFactory.create(MethodDeclarations.replicateAllMethod_id, l), false, true, 5000, false);
         }
         catch (Throwable t)
         {
            log.error("failed replicating " + l.size() + " elements in replication queue", t);
         }
      }
   }

   class MyTask extends TimerTask
   {
      public void run()
      {
         flush();
      }
   }
}